@@ -1,12 +1,11 @@
 import ast
-from scy import exceptions
 from typing import Any, Union
-from scy.tokenizer import Tokenizer
+
+from scy import exceptions
+from scy.tokens import (BINARY_OPERATORS, COMPARISON_OPERATORS,
+                        UNARY_OPERATORS, Token, TokenGroup, TokenType)
 from scy.utils import find_line
 
-from scy.tokens import BINARY_OPERATORS, COMPARISON_OPERATORS, TokenGroup, TokenType, Token, UNARY_OPERATORS
-
-
 ASSIGNABLES = (
     ast.Attribute,
     ast.Subscript,
@@ -79,7 +78,13 @@ class Parser:
             raise self.error(self.previous(), exceptions.INVALID_ASYNC % self.previous().lexeme)
 
     def statement(self, is_async: bool = False) -> list[ast.stmt]:
-        if self.match_(TokenType.FOR):
+        if self.match_(TokenType.IMPORT):
+            self.raise_if_async(is_async)
+            return [self.import_statement()]
+        elif self.match_(TokenType.FROM):
+            self.raise_if_async(is_async)
+            return [self.from_statement()]
+        elif self.match_(TokenType.FOR):
             return self.for_statement(is_async)
         elif self.match_(TokenType.IF):
             self.raise_if_async(is_async)
@@ -102,6 +107,58 @@ class Parser:
             raise self.error(self.peek(), exceptions.INVALID_ASYNC_EXPR)
         return [self.expression_statement()]
 
+    def import_statement(self) -> ast.stmt:
+        word = self.previous()
+        names = self.names_with_alias()
+        semi = self.consume(TokenType.SEMICOLON, "Expect ';' after import statement.")
+        return self.ast_token(names, klass=ast.Import, first=word, last=semi)
+
+    def from_statement(self) -> ast.stmt:
+        from_word = self.previous()
+        module, module_name_first, dotted_name_last = self.dotted_name('Expect module name.')
+        level = 0
+        for char in module:
+            if char == '.':
+                level += 1
+            else:
+                break
+        module = module[level:]
+        self.consume(TokenType.IMPORT, "Expect 'import' after module name.")
+        names = self.names_with_alias(allow_star=True)
+        semi = self.consume(TokenType.SEMICOLON, "Expect ';' after from..import statement.")
+        return self.ast_token(module, names, level, klass=ast.ImportFrom, first=from_word, last=semi)
+
+    def names_with_alias(self, allow_star: bool = False) -> list[ast.alias]:
+        paren = self.match_(TokenType.LEFT_PAREN)
+        names = [self.name_with_alias(allow_star)]
+        while self.match_(TokenType.COMMA):
+            names.append(self.name_with_alias(allow_star))
+        if paren:
+            self.consume(TokenType.RIGHT_PAREN, "Expect ')' after names.")
+        return names
+
+    def name_with_alias(self, allow_star: bool = False) -> ast.alias:
+        if allow_star and self.match_(TokenType.STAR):
+            return self.ast_token('*', None, klass=ast.alias, first=self.previous())
+        name, name_first, name_last = self.dotted_name('Expect name to import.')
+        if name.startswith('.'):
+            raise self.error(name_first, 'Cannot have relative name here.')
+        if self.match_(TokenType.AS):
+            asname, asname_first, asname_last = self.dotted_name('Expect alias name.')
+            return self.ast_token(name, asname,
+                                  klass=ast.alias, first=name_first, last=asname_last)
+        else:
+            return self.ast_token(name, None, klass=ast.alias, first=name_first, last=name_last)
+
+    def dotted_name(self, error: str) -> tuple[str, Token, Token]:
+        first = self.consume_any((TokenType.DOT, TokenType.IDENTIFIER), error)
+        result = first.lexeme
+        last = first
+        while self.match_(TokenType.DOT, TokenType.IDENTIFIER):
+            last = self.previous()
+            result += last.lexeme
+        return result, first, last
+
     def return_statement(self) -> ast.stmt:
         keyword = self.previous()
         if self.match_(TokenType.SEMICOLON):
@@ -219,6 +276,28 @@ class Parser:
             self.consume(end, error)
         return statement
 
+    def multi_param_list(self, tokens: dict[TokenType, TokenType],
+                         allow_empty: bool = True,
+                         error: str = 'Invalid list token %r.') -> list[ast.expr]:
+        if self.match_(*tokens.keys()):
+            ending = tokens[self.previous().type]
+        elif allow_empty:
+            ending = None
+        else:
+            raise self.error(self.peek(), error % self.peek().lexeme)
+        return self.param_list(ending, f'Expected {ending.name.lower()} to end list.')
+
+    def param_list(self, end_token: TokenType = None, error: str = '') -> list[ast.expr]:
+        if end_token is not None:
+            if self.match_(end_token):
+                return []
+        result = [self.expression(False)]
+        while not self.match_(TokenType.COMMA):
+            result.append(self.expression(False))
+        if end_token is not None:
+            raise self.error(self.peek(), error)
+        return result
+
     def optional_block(self, fill_empty: bool = True) -> list[ast.stmt]:
         if self.match_(TokenType.LEFT_BRACE):
             result = self.block()
@@ -471,7 +550,7 @@ class Parser:
 
     def consume_any(self, types: tuple[TokenType], message: str) -> Token:
         if self.match_(*types):
-            return self.peek()
+            return self.previous()
         raise self.error(self.peek(), message)
 
     def error(self, token: Token, message: str) -> SyntaxError:
